/******************************************************************************

  Copyright (c) 2003-2007 Turku PET Centre

  Program:     e7emhdr
  Description: Edit the ECAT 7.x mainheader.

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
  Place, Suite 330, Boston, MA 02111-1307 USA.

  Turku PET Centre hereby disclaims all copyright interest in the program.
  Juhani Knuuti
  Director, Professor
  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi/

  Modification history:
  0.1   2003-10-08 Vesa Oikonen
        First created.
  0.2   2004-08-15 VO
        Accepts options.
        Field names and values can be given in a file.
  0.2.1 2004-10-13 VO
        Sets Daylight saving time to unknown (-1) for Sun/Unix.
  0.3.0 2007-03-15 VO
        New libraries applied.
        Build information is printed in separate function.
        Accepts option --silent.
        Options -v, --version, --build, -test[=<level>], -h, --help.
        Lots of changes in info text.


******************************************************************************/
#define PROGRAM "e7emhdr 0.3.0  (c) 2003-2007 by Turku PET Centre"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
/*****************************************************************************/
#include "libtpcimgiov.h"
//#include "libtpcimgpv.h"
#include "libtpcmiscv.h"
#include "ecat7.h"
#include "ift.h"
/*****************************************************************************/
int TEST=0;
int SILENT=0;
/*****************************************************************************/
/* Local functions */
void print_usage();
void print_build();
int ecat7EditMHeader(ECAT7_mainheader *h, char *field, char *value);
/*****************************************************************************/

/*****************************************************************************/
/** Edit ECAT 7 main header. Returns 0, if ok, and 1 or 2, if field name or
    or value is invalid. */
int ecat7EditMHeader(ECAT7_mainheader *h, char *field, char *value)
{
  int ii, yy, mm, dd, ho, mi, se;
  struct tm *st;
  time_t timet;
  short int si;
  float f;

  si=atoi(value); ii=atoi(value); f=atof(value);
  if(strcmp(field, "magic_number")==0) {
    strncpy(h->magic_number, value, 14);
  } else if(strcmp(field, "original_file_name")==0) {
    strncpy(h->original_file_name, value, 32);
  } else if(strcmp(field, "sw_version")==0) {
    if(si<=0) return(2); else h->sw_version=si;
  } else if(strcmp(field, "system_type")==0) {
    if(si<0) return(2); else h->system_type=si;
  } else if(strcmp(field, "file_type")==0) {
    if(si<0) return(2); else h->file_type=si;
  } else if(strcmp(field, "serial_number")==0) {
    strncpy(h->serial_number, value, 10);
  } else if(strcmp(field, "scan_start_time")==0) {
    timet=time(NULL); st=localtime(&timet);
    if(sscanf(value, "%d-%d-%d %d:%d:%d", &yy, &mm, &dd, &ho, &mi, &se)!=6)
      return(2);
    st->tm_mday=dd; st->tm_mon=mm-1; st->tm_year=yy-1900;
    st->tm_hour=ho; st->tm_min=mi; st->tm_sec=se; st->tm_isdst=-1;
    h->scan_start_time=mktime(st);
  } else if(strcmp(field, "isotope_name")==0) {
    strncpy(h->isotope_name, value, 8);
  } else if(strcmp(field, "isotope_halflife")==0) {
    if(f<=1.0E-3) return(2); else h->isotope_halflife=f;
  } else if(strcmp(field, "radiopharmaceutical")==0) {
    strncpy(h->radiopharmaceutical, value, 32);
  } else if(strcmp(field, "gantry_tilt")==0) {
    h->gantry_tilt=f;
  } else if(strcmp(field, "gantry_rotation")==0) {
    h->gantry_rotation=f;
  } else if(strcmp(field, "bed_elevation")==0) {
    h->bed_elevation=f;
  } else if(strcmp(field, "intrinsic_tilt")==0) {
    h->intrinsic_tilt=f;
  } else if(strcmp(field, "wobble_speed")==0) {
    h->wobble_speed=si;
  } else if(strcmp(field, "transm_source_type")==0) {
    h->transm_source_type=si;
  } else if(strcmp(field, "distance_scanned")==0) {
    h->distance_scanned=f;
  } else if(strcmp(field, "transaxial_fov")==0) {
    h->transaxial_fov=f;
  } else if(strcmp(field, "angular_compression")==0) {
    h->angular_compression=si;
  } else if(strcmp(field, "coin_samp_mode")==0) {
    h->coin_samp_mode=si;
  } else if(strcmp(field, "axial_samp_mode")==0) {
    h->axial_samp_mode=si;
  } else if(strcmp(field, "ecat_calibration_factor")==0) {
    h->ecat_calibration_factor=f;
  } else if(strcmp(field, "calibration_units")==0) {
    h->calibration_units=si;
  } else if(strcmp(field, "calibration_units_label")==0) {
    h->calibration_units_label=si;
  } else if(strcmp(field, "compression_code")==0) {
    h->compression_code=si;
  } else if(strcmp(field, "study_type")==0) {
    strncpy(h->study_type, value, 12);
  } else if(strcmp(field, "patient_id")==0) {
    strncpy(h->patient_id, value, 16);
  } else if(strcmp(field, "patient_name")==0) {
    strncpy(h->patient_name, value, 32);
  } else if(strcmp(field, "patient_sex")==0) {
    h->patient_sex=value[0];
  } else if(strcmp(field, "patient_dexterity")==0) {
    h->patient_dexterity=value[0];
  } else if(strcmp(field, "patient_age")==0) {
    h->patient_age=f;
  } else if(strcmp(field, "patient_height")==0) {
    h->patient_height=f;
  } else if(strcmp(field, "patient_weight")==0) {
    h->patient_weight=f;
  } else if(strcmp(field, "patient_birth_date")==0) {
    timet=time(NULL); st=localtime(&timet);
    if(sscanf(value, "%d-%d-%d", &yy, &mm, &dd)!=3) return(2);
    st->tm_mday=dd; st->tm_mon=mm-1; st->tm_year=yy-1900;
    st->tm_hour=0; st->tm_min=0; st->tm_sec=0; st->tm_isdst=-1;
    h->patient_birth_date=mktime(st);
  } else if(strcmp(field, "physician_name")==0) {
    strncpy(h->physician_name, value, 32);
  } else if(strcmp(field, "operator_name")==0) {
    strncpy(h->operator_name, value, 32);
  } else if(strcmp(field, "study_description")==0) {
    strncpy(h->study_description, value, 32);
  } else if(strcmp(field, "acquisition_type")==0) {
    h->acquisition_type=si;
  } else if(strcmp(field, "patient_orientation")==0) {
    h->patient_orientation=si;
  } else if(strcmp(field, "facility_name")==0) {
    strncpy(h->facility_name, value, 20);
  } else if(strcmp(field, "num_planes")==0) {
    h->num_planes=si;
  } else if(strcmp(field, "num_frames")==0) {
    h->num_frames=si;
  } else if(strcmp(field, "num_gates")==0) {
    h->num_gates=si;
  } else if(strcmp(field, "num_bed_pos")==0) {
    h->num_bed_pos=si;
  } else if(strcmp(field, "init_bed_position")==0) {
    h->init_bed_position=f;
  } else if(strcmp(field, "bed_position")==0) {
    sscanf(value, "%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
      h->bed_position+0, h->bed_position+1, h->bed_position+2,
      h->bed_position+3, h->bed_position+4, h->bed_position+5,
      h->bed_position+6, h->bed_position+7, h->bed_position+8,
      h->bed_position+9, h->bed_position+10, h->bed_position+11,
      h->bed_position+12, h->bed_position+13, h->bed_position+14
    );
  } else if(strcmp(field, "plane_separation")==0) {
    h->plane_separation=f;
  } else if(strcmp(field, "lwr_sctr_thres")==0) {
    h->lwr_sctr_thres=si;
  } else if(strcmp(field, "lwr_true_thres")==0) {
    h->lwr_true_thres=si;
  } else if(strcmp(field, "upr_true_thres")==0) {
    h->upr_true_thres=si;
  } else if(strcmp(field, "user_process_code")==0) {
    strncpy(h->user_process_code, value, 10);
  } else if(strcmp(field, "acquisition_mode")==0) {
    h->acquisition_mode=si;
  } else if(strcmp(field, "bin_size")==0) {
    h->bin_size=f;
  } else if(strcmp(field, "branching_fraction")==0) {
    h->branching_fraction=f;
  } else if(strcmp(field, "dose_start_time")==0) {
    timet=time(NULL); st=localtime(&timet);
    if(sscanf(value, "%d-%d-%d %d:%d:%d", &yy, &mm, &dd, &ho, &mi, &se)!=6)
      return(2);
    st->tm_mday=dd; st->tm_mon=mm-1; st->tm_year=yy-1900;
    st->tm_hour=ho; st->tm_min=mi; st->tm_sec=se; st->tm_isdst=-1;
    h->dose_start_time=mktime(st);
  } else if(strcmp(field, "dosage")==0) {
    h->dosage=f;
  } else if(strcmp(field, "well_counter_corr_factor")==0) {
    h->well_counter_corr_factor=f;
  } else if(strcmp(field, "data_units")==0) {
    strncpy(h->data_units, value, 32);
  } else if(strcmp(field, "septa_state")==0) {
    h->septa_state=si;
  } else
    return(1);

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/*
 *  main()
 */
int main(int argc, char *argv[])
{
  int                 ai, li, ret, change_nr=0;
  char               *cptr, petfile[FILENAME_MAX], iftfile[FILENAME_MAX];
  char                keyname[FILENAME_MAX], eqsign[FILENAME_MAX],
                      keyvalue[FILENAME_MAX];
  FILE               *fp;
  ECAT7_mainheader    main_header;
  IFT                 ift;


  /*
   *  Get arguments
   */
  if(argc==1) {print_usage(); return(1);}
  petfile[0]=iftfile[0]=(char)0;
  keyname[0]=eqsign[0]=keyvalue[0]=(char)0;
  iftInit(&ift);
  /* Options */
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') {
    cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(cptr==NULL) continue;
    if(strncasecmp(cptr, "TEST", 4)==0) {
      cptr+=4; if(*cptr=='=') cptr++; TEST=atoi(cptr); if(TEST<1) TEST=1;
      continue;
    } else if(strcasecmp(cptr, "V")==0 || strcasecmp(cptr, "VERSION")==0 ||
              strcasecmp(cptr, "BUILD")==0) {
      print_build();
      return(0);
    } else if(strcasecmp(cptr, "H")==0 || strcasecmp(cptr, "HELP")==0) {
      print_usage(); return(0);
    } else if(strcasecmp(cptr, "SILENT")==0) {
      SILENT=1; continue;
    }
    fprintf(stderr, "Error: invalid option %s\n", argv[ai]);
    return(1);
  } else { /* arguments */
    if(!petfile[0]) {
      strcpy(petfile, argv[ai]); continue;
    } else if(!keyname[0]) {
      strcpy(keyname, argv[ai]); continue;
    } else if(!eqsign[0]) {
      strcpy(eqsign, argv[ai]); continue;
    } else if(!keyvalue[0]) {
      strcpy(keyvalue, argv[ai]); continue;
    }
    fprintf(stderr, "Error: invalid argument %s\n", argv[ai]);
    return(1);
  }
  if(!keyname[0]) {print_usage(); return(1);}
  if(!eqsign[0]) { /* then filename was given instead of keyname */
    strcpy(iftfile, keyname); keyname[0]=(char)0;
  } else {
    if(strcmp(eqsign, ":=")!=0) { /* check if eqsign is ok */
      fprintf(stderr, "Error: use ' := ' to separate field name and value.\n");
      return(1);
    }
  }
  if(keyname[0] && !keyvalue[0]) { /* Both keyname and value must be given */
    print_usage(); return(1);
  } else {
    iftPut(&ift, keyname, keyvalue, NULL);
  }
  if(TEST>2) ECAT7_TEST=TEST-2;
  if(TEST>2) IFT_TEST=TEST-2;
  if(TEST) printf("********************* Test run *********************\n");
  if(TEST) {
    printf("TEST := %d\n", TEST);
    printf("SILENT := %d\n", SILENT);
    printf("petfile := %s\n", petfile);
    printf("iftfile := %s\n", iftfile);
    if(keyname[0]) printf("%s := %s\n", keyname, keyvalue);
  }


  /*
   *  Read key names and values from file, if necessary
   */
  if(iftfile[0]) {
    if(TEST) printf("reading %s\n", iftfile);
    ret=iftRead(&ift, iftfile, 1);
    if(ret) {
      fprintf(stderr, "Error (%d): %s\n", ret, ift.status);
      iftEmpty(&ift); return(2);
    }
  }
  if(TEST>10) iftWrite(&ift, "stdout");


  /*
   *  Read main header
   */
  /* Open input ECAT file */
  if(TEST) printf("reading %s\n", petfile);
  if((fp=fopen(petfile, "rb")) == NULL) {
    fprintf(stderr, "Error: cannot open file %s\n", petfile); 
    iftEmpty(&ift); return(2);
  }
  /* Read the header */
  if((ret=ecat7ReadMainheader(fp, &main_header))) {
    fprintf(stderr, "Error (%d): cannot read main header.\n", ret);
    iftEmpty(&ift); fclose(fp); return(3);
  }
  /* Close ECAT file */
  fclose(fp);
  if(TEST>9) ecat7PrintMainheader(&main_header, stdout);
#if(0) // do not check, otherwise cannot edit empty header
  /* Check file format */
  if(strncmp(main_header.magic_number, ECAT7V_MAGICNR, 7)!=0) {
    fprintf(stderr, "Error: %s is not an ECAT 7.x file.\n", petfile);
    iftEmpty(&ift); return(3);
  }
#endif  


  /*
   *  Change the fields
   */
  for(li=0, change_nr=0; li<ift.keyNr; li++) {
    if(TEST>1) iftWriteItem(&ift, li, stdout);
    /* Change field value */
    ret=ecat7EditMHeader(&main_header, ift.item[li].key, ift.item[li].value);
    if(ret==1) {
      fprintf(stderr, "  warning: invalid field name: %s\n", ift.item[li].key);
    } else if(ret==2) {
      fprintf(stderr, "  warning: invalid field value: %s\n", ift.item[li].value);
    } else change_nr++;
  } /* next field */
  if(TEST>3) ecat7PrintMainheader(&main_header, stdout);
  iftEmpty(&ift);
  if(change_nr==0) {
    fprintf(stderr, "Error: no fields were changes.\n"); 
    return(8);
  }


  /*
   *  Write main header
   */
  /* Open input CTI file */
  if((fp=fopen(petfile, "r+b")) == NULL) {
    fprintf(stderr, "Error: cannot open %s for write.\n", petfile); 
    return(11);
  }
  /* Read header */
  if((ret=ecat7WriteMainheader(fp, &main_header))) {
    fprintf(stderr, "Error (%d): cannot write main header.\n", ret);
    fclose(fp); return(12);
  }
  /* Close CTI file */
  fclose(fp);

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/*
 *   print_usage()
 */
void print_usage()
{
  int i=0;
  static char *info[] = {
  " ",
  PROGRAM,
  " ",
  "Edits the information in the main header of an ECAT 7 file.",
  " ",
  "Usage: e7emhdr [Options] <ECAT file> <field_name> := <field value> ",
  "   or: e7emhdr [Options] <ECAT file> <header file>",
  " ",
  "Options:",
  " -h or --help",
  "     Print this message and exit",
  " -v, --version or --build",
  "     Print software build information and exit",
  " --silent",
  "     Program works silently, printing only error and warning messages.",
  " ",
  "Header file can contain one or more mainheader field names and values",
  "in interfile-like format, e.g.",
  "  patient_name := Einstein Albert",
  "  radiopharmaceutical := O-15 WATER",
  "  study_description := brain scan",
  "  facility_name := Turku PET Centre",
  " ",
  "Examples:",
  "  e7emhdr s2345dy1.v patient_name := \"Albert Einstein\"",
  "  e7emhdr s2345dy1.v s2345dy1.header",
  " ",
  "See also: lmhdr, lshdr, e7evhdr, lmlist, ecatunit",
  " ",
  "Keywords: ECAT7, image, header, tools",
  " ",
  "This program comes with ABSOLUTELY NO WARRANTY. This is free software, and",
  "you are welcome to redistribute it under GNU General Public License.",
  " ",
  0};
  while(info[i]!=0) printf("  %s\n", info[i++]);
}
/*****************************************************************************/

/*****************************************************************************/
/**
 * Prints build information to stdout.
 */
void print_build(){

  printf("\n %s\n", PROGRAM);
  printf("\n Build %s %s\n\n",__DATE__,__TIME__);
#if defined(__FILE__) && defined(__TIMESTAMP__)
  printf(" File %s last edited on %s\n", __FILE__, __TIMESTAMP__);
#endif
  printf(" _______________________________________________\n");

//  libtpcimgp_print_build(stdout);
//  printf(" _______________________________________________\n");
  libtpcimgio_print_build(stdout);
  printf(" _______________________________________________\n");
  libtpcmisc_print_build(stdout);
  printf(" _______________________________________________\n");

#if defined(__STDC_VERSION__)
  printf(" Version of C: %ld\n", __STDC_VERSION__);
  printf(" _______________________________________________\n\n");
#endif
#if defined(__STDC_ISO_10646__)
  printf(" Compiler ISO/IEC 10646 specification: %ld\n", __STDC_ISO_10646__);
  printf(" _______________________________________________\n\n");
#endif
#if defined(__GNUC__) && defined(__VERSION__)
  printf(" GNU C version: %s\n", __VERSION__);
  printf(" _______________________________________________\n\n");
#endif
}
/*****************************************************************************/

/*****************************************************************************/
